/** 
 *   Site testing for SKA candidate.
 *
 *   (c) Copyright 2004 Federico Bareilles <fede@iar.unlp.edu.ar>,
 *   Instituto Argentino de Radio Astronomia (IAR).
 *
 *   The GPIB implementation for spectral analyzer is initially writen
 *   by Cecilia Morras (on LavWindows).
 * 
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the License, or (at your option) any later version.
 *     
 *   The author does NOT admit liability nor provide warranty for any
 *   of this software. This material is provided "AS-IS" in the hope
 *   that it may be useful for others.
 *
 *   TODO:
 *        !) Comunicacion con el analizador espectral.
 *        *) llaves de RF via paralelo:
 *           - falta descripcion de hardware.
 *           ! falta implementacion.
 *           - falta probarlo.
 *        !) comunicacion con el rotador:
 *           ! falta descripcion de hardware.
 *           ! falta implementacion.
 *           - falta probarlo.
 *           ! Grilllado para el rotador.
 *        !) Version.
 *        !) Estructura de directorios para los archivos.
 *
 *
 **/

#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <getopt.h>
#include <syslog.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/file.h>

#include "version.h"
#include "my_stdio.h"
#include "toska.h"
#include "observ_table.h"
#include "observation.h"
#include "analyzer.h"
#include "switch_rf.h"
#include "rotator.h"
#include "plot.h"
#include "thermometer.h"

/* Flags: */
#ifdef NODAEMON
  int daemon_flag       = 0;
#else
  int daemon_flag       = 1;
#endif
int rotator_flag = 1;
int thermometer_flag = 1;
int plot_flag = 0;
int current_flag = 1;
int restart_flag = 0;


const char *name_definition = {"Tiny Observatory for SKA"};
const char *help_text[] ={
#       ifdef VERBOSE_MODE
	"  -v\t\t - verbose mode (incremental).",
#       endif
	"  -n, --nodaemon - no run as a daemon.",
	"  -l,[--log]\t - write events to syslog.",
	"  -b\t\t - set base directoy (def.: "BASE_DIR")",
	"  -d\t\t - set data directoy (def.: base_dir/"DATA_DIR")",
	"  -m\t\t - set modes definition file (def.: base_dir/"MODES_FILE")",
	"  -s\t\t - set sequence file (def.: base_dir/"SEQUENCE_FILE")",
	"  --swrf\t - set rf switch port in exadecimal.",
	"  --norot\t - Ignored rotator.",
	"  --nother\t - Ignored thermometer.",
	"  -g\t\t - plot all spectrum (implies -n).",
	"  -V\t\t - show program version.",
	"  --help\t - this help !!!.",
	""
};

void version(char *program_name)
{
	fprintf(stderr,"%s: ", name_definition);
	fprintf(stderr,"%s version %s\n", program_name, VERSION);
}

void usage(char *program_name)
{
	version(program_name);
	fprintf(stderr,
#               ifdef VERBOSE_MODE
"Usage: \t%s [-vnlbdmsgV][--nodaemon][--log][--swrf port][--norot][--nother][--help]\n",
#                else
 "Usage: \t%s [-nlbdmsgV][--nodaemon][--log][--swrf port][--norot][--nother][--help]\n",
#               endif
		program_name);
}


void help()
{
	int i=0;

	while((char) *help_text[i] != 0)
		fprintf(stderr,"%s\n",help_text[i++]);
	fprintf(stderr,"\n");
}


void close_log(void)
{
	SYSLOG(LOG_INFO, "Program terminated.");
}

void cbrk_handler(int s)
{
	antenna_switchs_close();
	SYSLOG(LOG_INFO, "receiving signal %d.", s);
	exit(1);
}

void hup_handler(int s)
{
	SYSLOG(LOG_INFO, "receiving signal Hangup.");
	restart_flag = 1;
}


int main(int argc, char **argv)
{
	int an_dev;
	int i;
	char base_dir[_POSIX_PATH_MAX];
	char data_dir[_POSIX_PATH_MAX];
	char modes_file[_POSIX_PATH_MAX];
	char sequence_file[_POSIX_PATH_MAX];
	char *program_name;
	int switch_rf_port = SWITCH_RF_PORT;
#ifndef NODAEMON
	int pid;
	int fd_pid;
	char pid_file[_POSIX_PATH_MAX];
#endif
	struct modes *obs, *obsp;
	struct sequence *seq, *seqp;
	struct obs_setup *s;
	char *cp;
	int op;
	static struct option  const  longopts[] =
	{
		{"help", 0, 0, 101},
		{"nodaemon", 0, 0, 'n'},
		{"log", 0, 0, 'l'},
		{"swrf", 1, 0, 201},
		{"norot", 0, 0, 202},
		{"nother", 0, 0, 203},
		{0,  0,  0,  0}
	};

	base_dir[0] = 0;
	data_dir[0] = 0;
	modes_file[0] = 0;
	sequence_file[0] = 0;

	/* who is my name? */
	if ((cp = strrchr(argv[0], '/')) != NULL) program_name  = cp + 1;
	else program_name  = argv[0];
  
	verbose_flag = 0;
	opterr =  0;
	while  ((op = getopt_long(argc, argv, "Vvnlb:d:m:fs:g",
				  longopts, 0)) != EOF)
		switch( op ){
		case 'v':
			INCREMENT_VERBOSE_FLAG();
			/* verbose_flag = (verbose_flag << 1) | 1; */
			break;
		case 101:
			usage(program_name);
			help();
			exit(-1);
		case 'n':
			daemon_flag = 0;
			break;
		case 'l':
			syslog_flag = 1;
			break;
		case 'b':
			sprintf(base_dir, "%s", optarg);
			if(base_dir[strlen(base_dir) - 1] == '/' )
				base_dir[strlen(base_dir) - 1] = 0;
			break;
		case 'd':
			sprintf(data_dir, "%s", optarg);
			if(data_dir[strlen(data_dir) - 1] == '/' )
				data_dir[strlen(data_dir) - 1] = 0;
			break;
		case 'f':
			usage(program_name);
			help();
			printf("*** Option -f deprecated, see help. ***\n\n");
			exit(-1);
		case 'm':
			sprintf(modes_file, "%s", optarg);
			break;
		case 201:
			switch_rf_port = strtol(optarg, (char **)NULL, 16);
			break;
		case 202:
			rotator_flag = 0;
			break;
		case 203:
			thermometer_flag = 0;
			break;
		case 's':
			sprintf(sequence_file, "%s", optarg);
			break;
		case 'g':
			daemon_flag = 0;
			plot_flag = 1;
			break;
		case 'V':
			version(program_name);
			exit(-1);
			break;
		default:
			usage(program_name);
			exit(-1);
		}
#       ifndef NODAEMON
	if(daemon_flag){
		/* Ok, a daemond */
		verbose_flag = 0;
		if ( (pid=fork()) != 0 ) {
			sprintf(pid_file, "/var/run/%s.pid", program_name);
			fd_pid = open(pid_file, O_WRONLY|O_CREAT|O_SYNC,
				      S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH);
			fd_printf(fd_pid,"%d\n", pid);
			close(fd_pid);
			exit(0);
		}
	}
#       endif
	
	if (base_dir[0] == 0) {
		sprintf(base_dir, "%s", BASE_DIR);
	}
	if (data_dir[0] == 0) {
		sprintf(data_dir, "%s/%s", base_dir, DATA_DIR);
	}
	if (modes_file[0] == 0) {
		sprintf(modes_file, "%s/%s", base_dir, MODES_FILE);
	}
	if (sequence_file[0] == 0) {
		sprintf(sequence_file, "%s/%s", base_dir, SEQUENCE_FILE);
	}

	if(syslog_flag){
		openlog(program_name, LOG_PID, LOG_DAEMON );
		atexit(close_log);
		SYSLOG(LOG_INFO, "Starting %s...", program_name);
		SYSLOG(LOG_INFO, "using base dir:   %s", base_dir);
		SYSLOG(LOG_INFO, "using data dir:   %s", data_dir);
		SYSLOG(LOG_INFO, "using modes file: %s", modes_file);
		SYSLOG(LOG_INFO, "using seq. file:  %s", sequence_file);
		SYSLOG(LOG_INFO, "rot. flag = %d, plot flag = %d",
		      rotator_flag, plot_flag);
	}

	if(signal(SIGINT, cbrk_handler) == SIG_ERR) {
		SYSLOG(LOG_ERR, "Signal Failed (SIGINT).");
		exit(1);
	}
	if(signal(SIGTERM, cbrk_handler) == SIG_ERR) {
		SYSLOG(LOG_ERR, "Signal Failed (SIGTERM).");
		exit(1);
	}
	if(signal(SIGHUP, hup_handler) == SIG_ERR) {
		SYSLOG(LOG_ERR, "Signal Failed (SIGHUP).");
		exit(1);
	}
	if(plot_flag) 
		open_plot();

	s = (struct obs_setup *) malloc(sizeof(struct obs_setup));
	if( s == NULL){
		SYSLOG(LOG_ERR, "Canot alocated memory (struct obs_setup).");
		exit(1);
	}
	if(syslog_flag){
		SYSLOG(LOG_NOTICE, "Startup succeeded.");
		
	}
	s->base_dir = base_dir;
	s->data_dir = data_dir;
	s->azimuth = -1;
	s->naxis1 = -1;
	s->polarization = 0;

	i = 0;
	do {
		an_dev = open_analyzer(AN_PADDR);
		if (an_dev < 0) {
			sleep(1);
		}
		i++;
	}while(an_dev < 0 && i < AN_RETRY_TIME_OUT);
	
	if (rotator_flag) {
		if ((s->fd_rot = rotator_setup(ROTATOR_DEVICE)) < 0) {
			SYSLOG(LOG_CRIT, "Canot open rotator device (%s).",
			       ROTATOR_DEVICE);
		}
	}

	if (thermometer_flag) {
		if ((s->fd_ther = thermometer_open(THERMOMETER_DEVICE)) < 0) {
			SYSLOG(LOG_CRIT, "Canot open thermometer device (%s).",
			       THERMOMETER_DEVICE);
		}
	}else{
		s->fd_ther = -1;
	}

	if (antenna_switchs_open(switch_rf_port) != 0) {
		SYSLOG(LOG_CRIT, "Cannot acces rf switch port.");
#ifndef DEMO_MODE
		exit(1);
#endif
	}

	do {
		restart_flag = 0;
		PVERB(PINFO,"Reading modes file: %s\n", modes_file);
		obs = obsp = get_modes_table(modes_file);
		if( obs == NULL){
			SYSLOG(LOG_ERR, "Error opening modes file: %s.",
			       modes_file );
			exit(1);
		}
		s->obs = obs;
		
		PVERB(PINFO,"Reading sequence file: %s\n", sequence_file);
		seq = seqp = get_sequence_table(sequence_file);
		if( seq == NULL){
			SYSLOG(LOG_ERR, "Error opening sequence file: %s.",
			       sequence_file );
			exit(1);
		}
		
		if (an_dev >= 0) {
			SYSLOG(LOG_INFO, "start observation.");
			observ_sequences(an_dev, s, seq, obs);
		}
		
		free_modes_table(obs);
		free_sequence_table(seq);
	} while(restart_flag);
	
	antenna_switchs_close();
	if (thermometer_flag)
		thermometer_close(s->fd_ther);
	if (rotator_flag)
		rotator_close(s->fd_rot);
	free(s);
	if (plot_flag)
		close_plot();

	SYSLOG(LOG_INFO, "stop observation.");

	return 0;
}
